# -*- coding:utf-8 -*-
import pytest
from .schemas import (
    Point,
    Pair,
    Plot
)


def test_point__name():
    pt = Point(x="10", y="20", z=None)
    assert pt.x.name == "x"


def test_point__value():
    pt = Point(x="10", y="20", z=None)
    assert pt.x.value == "10"


def test_point__validation():
    pt = Point(x="10", y="20", z=None)
    data = pt.validate()
    assert data["x"] == 10
    assert pt.x.raw == "10"


def test_point__validation_failure():
    from tinyschema import Failure
    pt = Point(x="10", y="20", z=None)
    pt.z.options["required"] = True

    with pytest.raises(Failure):
        pt.validate()


def test_point_validator__failure():
    from tinyschema import Failure
    from tinyschema.datavalidation import multi, ValidationObject, Invalid

    class EqualValidation(ValidationObject):
        @multi(["x", "y"])
        def equals(self, x, y):
            if x != y:
                raise Invalid("oops")
    pt = Point(x="10", y="20", z=None)
    with pytest.raises(Failure):
        EqualValidation()(pt)


def test_point_validator__success():
    from tinyschema.datavalidation import multi, ValidationObject, Invalid

    class EqualValidation(ValidationObject):
        @multi(["x", "y"])
        def equals(self, x, y):
            if x != y:
                raise Invalid("oops")
    pt = Point(x="10", y="10", z=None)
    data = EqualValidation()(pt)
    assert data["x"] == 10


def test_point_validator__skip():
    from tinyschema.datavalidation import multi, ValidationObject, Invalid

    class EqualValidation(ValidationObject):
        @multi(["z", "y"])
        def equals(self, x, y):
            if x != y:
                raise Invalid("oops")
    pt = Point(x="10", y="20", z=None)
    data = EqualValidation()(pt)
    assert data["x"] == 10


def test_pair__name():
    pair = Pair(l=Point(x="10", y="20"),
                r=Point(x="10", y="20"))

    assert pair.l.x.name == "x"
    assert pair.r.x.name == "x"


def test_pair__value():
    pair = Pair(l=Point(x="10", y="20"),
                r=Point(x="100", y="200"))

    assert pair.l.x.value == "10"
    assert pair.r.x.value == "100"


def test_pair__value2():
    pair = Pair(l=dict(x="10", y="20"),
                r=dict(x="100", y="200"))

    assert pair.l.x.value == "10"
    assert pair.r.x.value == "100"


def test_pair__label():
    pair = Pair(l=dict(x="10", y="20"),
                r=dict(x="100", y="200"))

    assert pair.l.label == "l"
    assert pair.r.label == "r"


def test_pair__validation():
    pair = Pair(l=dict(x="10", y="20"),
                r=dict(x="100", y="200"))
    data = pair.validate()

    assert data["l"]["x"] == 10
    assert pair.l.x.raw == "10"
    assert data["r"]["x"] == 100
    assert pair.r.x.raw == "100"


def test_pair__validation_failure():
    from tinyschema import Failure
    pair = Pair(l=dict(x="10", y="20"),
                r=dict(x="100", y="200"))
    pair.l.z.options["required"] = True

    with pytest.raises(Failure):
        pair.validate()


def test_plot__name():
    plot = Plot(ps=[{"x": "10", "y": "20"}])
    assert plot.ps[0].x.name == "x"


def test_plot__value():
    plot = Plot(ps=[{"x": "10", "y": "20"}])
    assert plot.ps[0].x.value == "10"


def test_plot__value2():
    plot = Plot(ps=[Point(**{"x": "10", "y": "20"})])
    assert plot.ps[0].x.value == "10"


def test_plot__validation():
    plot = Plot(ps=[Point(**{"x": "10", "y": "20"})])
    data = plot.validate()

    assert data["ps"][0]["x"] == 10


def test_plot__validation2():
    from tinyschema import Failure
    plot = Plot(ps=[Point(**{"x": "10", "y": "20"})])
    plot.ps[0].z.options["required"] = True

    with pytest.raises(Failure):
        plot.validate()



# -*- coding:utf-8 -*-
import logging
logger = logging.getLogger(__name__)
import sys
from functools import partial
from collections import defaultdict
from . import Failure


def is_validator(v):
    return hasattr(v, "_v_count")


class Counter(object):
    def __init__(self):
        self.i = 0

    def __call__(self):
        self.i += 1
        return self.i

counter = Counter()


class Multi(object):
    def __init__(self, names, method, msg=None, position=None):
        self.names = names
        self.method = method
        self._v_count = counter()
        self.msg = msg
        self.position = position

    def __call__(self, parent, params):
        if all(params.get(name) is not None for name in self.names):
            return self.method(parent, *(params[name] for name in self.names))


class Matched(object):
    def __init__(self, names, method, msg=None):
        self.names = names
        self.method = method
        self._v_count = counter()
        self.msg = msg

    def __call__(self, parent, params):
        if any(params.get(name) is not None for name in self.names):
            return self.method(parent, [(name, params[name]) for name in self.names if params.get(name) is not None])


class Convert(object):
    def __init__(self, names, method, msg=None):
        self.names = names
        self.method = method
        self._v_count = counter()
        self.msg = msg

    def __call__(self, parent, params):
        if not self.names:
            return self.method(parent, params)
        else:
            if all(params.get(name) is not None for name in self.names):
                return self.method(parent, params)


class Container(object):
    def __init__(self, name, cls):
        self.names = [name]  # for common interface
        self.name = name
        self.validators = [v for v in cls.__dict__.values() if is_validator(v)]
        self.validators.sort(key=lambda o: o._v_count)
        self._v_count = counter()
        self.msg = None

        self.child = None

    def __call__(self, parent, params):
        if self.name in params:
            target = params[self.name]
            for v in self.validators:
                self.child = v
                v(parent, target)

    @property
    def position(self):
        r = [self.name]
        if self.child:
            r.append(getattr(self.child, "position") or self.child.names[0])
        return r


class Collection(object):
    def __init__(self, name, cls):
        self.names = [name]  # for common interface
        self.name = name
        self.validators = [v for v in cls.__dict__.values() if is_validator(v)]
        self.validators.sort(key=lambda o: o._v_count)
        self._v_count = counter()
        self.msg = None

        self.index = None
        self.child = None

    def __call__(self, parent, params):
        if self.name in params:
            targets = params[self.name]
            for i, target in enumerate(targets):
                for v in self.validators:
                    self.index = i
                    self.child = v
                    v(parent, target)

    @property
    def position(self):
        r = [self.name]
        if self.child:
            r.append(self.index)
            r.append(getattr(self.child, "position") or self.child.names[0])
        return r


def single(name, msg=None):
    return partial(Multi, [name], msg=msg)


def multi(names, msg=None, position=None):
    assert isinstance(names, (list, tuple))
    return partial(Multi, names, msg=msg, position=position)


def convert(names=None, msg=None):
    return partial(Convert, names, msg=msg)


def matched(names, msg=None):
    assert isinstance(names, (list, tuple))
    return partial(Matched, names, msg=msg)


def collection(name):
    return partial(Collection, name)


def container(name):
    return partial(Container, name)


def share(*factories):
    cls_env = sys._getframe(1).f_locals

    def wrapper(method):
        validations = []
        for f in factories:
            name = "{}{}".format(method.__name__, counter())
            v = f(method)
            cls_env[name] = v
            validations.append(v)
        return validations
    return wrapper


class Invalid(Exception):
    def __init__(self, msg, position=None):
        self.msg = msg
        self.position = position


class Interrupt(Exception):
    def __init__(self, msg, position=None):
        self.msg = msg
        self.position = position


class ValidationObjectMeta(type):
    def __new__(self, name, bases, attrs):
        validators = [v for v in attrs.values() if is_validator(v)]
        validators.sort(key=lambda o: o._v_count)
        attrs["validators"] = validators

        def __call__(self, ob):
            status = True
            params, errors = self.configure(ob)
            if errors:
                return self.on_failure(ob, params, errors)

            try:
                for v in self.validators:
                    try:
                        v(self, params)
                    except self.Exception as e:
                        self.catch_error(params, errors, v, e)
                        status = False
            except self.Interrupt as e:
                self.catch_error(params, errors, v, e)
                return self.on_failure(ob, params, errors)

            if status:
                return self.on_success(ob, params)
            else:
                return self.on_failure(ob, params, errors)

        attrs["__call__"] = __call__
        attrs["validate"] = __call__
        return super().__new__(self, name, bases, attrs)


class _ValidationObject(object):
    def configure(self, schema):
        try:
            return schema.validate(), defaultdict(list)
        except Failure as e:
            return {}, e.errors

    def catch_error(self, params, errors, validation, e):
        position = getattr(e, "position", None)
        if position is None:
            position = getattr(validation, "position", None) or validation.names[0]

        if validation.msg is None:
            msg = e.args[0]
        elif callable(validation.msg):
            msg = validation.msg(**params)
        else:
            msg = validation.msg

        if not isinstance(position, (list, tuple)):
            errors[position].append(msg)
        else:
            target = errors
            for i, k in enumerate(position[:-1]):
                if k not in target:
                    if isinstance(position[i + 1], int):
                        target[k] = [defaultdict(list) for i in range(position[i + 1] + 1)]
                    else:
                        target[k] = defaultdict(list)
                try:
                    target = target[k]
                except IndexError:
                    for i in range(k + 1):
                        target[i] = defaultdict(list)
            target[position[-1]].append(msg)

    def on_success(self, _, params):
        return params

    def on_failure(self, _, params, errors):
        raise Failure(errors)

ValidationObject = ValidationObjectMeta("ValidationObject", (_ValidationObject, ), {"Exception": Invalid, "Interrupt": Interrupt})



import sys

PY3 = sys.version_info[0] == 3

if PY3:  # pragma: no cover
    string_types = str,
    text_type = str
else:  # pragma: no cover
    string_types = basestring,
    text_type = unicode


def text_(s, encoding='latin-1', errors='strict'):
    """ If ``s`` is an instance of ``bytes``, return ``s.decode(encoding,
    errors)``, otherwise return ``s``"""
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    return s  # pragma: no cover

if PY3:  # pragma: no cover
    def is_nonstr_iter(v):
        if isinstance(v, str):
            return False
        return hasattr(v, '__iter__')
else:  # pragma: no cover
    def is_nonstr_iter(v):
        return hasattr(v, '__iter__')

try:
    xrange = xrange
except NameError:  # pragma: no cover
    xrange = range


try:
    from cPickle import loads, dumps, HIGHEST_PROTOCOL
except ImportError:  # pragma: no cover
    from pickle import loads, dumps, HIGHEST_PROTOCOL

if PY3:
    FileNotFoundError = FileNotFoundError
else:
    FileNotFoundError = IOError



# -*- coding:utf-8 -*-
import pytest
from .schemas import (
    Point,
    Pair,
    Plot
)

candidates = [
    (Point, "object"),
    (Pair, "object"),
    (Plot, "object"),
    (Point.x, "atom"),
    (Pair.l, "object"),
    (Plot.ps, "array")
]


@pytest.mark.parametrize("target, detected", candidates)
def test_object(target, detected):
    from tinyschema.inspection import type_of
    assert type_of(target) == detected



# -*- coding:utf-8 -*-
import logging
logger = logging.getLogger(__name__)


def iterate_params(params):
    if not hasattr(params, "keys"):
        for k, v in params:
            yield k, v
    elif "order" in params:
        for k in params["order"]:
            yield k, params[k]
    else:
        for k, v in params.items():
            yield k, v


class Mapper(object):
    def __init__(self, schema, transformer):
        self.schema = schema
        self.transformer = transformer

    def __call__(self, name, params):
        schema = self.schema(**params)
        parsed = schema.validate()
        return self.transformer(name, schema, parsed)


def default_iterator(triple_list):
    for name, typename, values in triple_list:
        yield name, typename, values


class MapperFamily(object):
    def __init__(self, aggregate, iterator=default_iterator):
        self.families = {}
        self.aggregate = aggregate
        self.iterator = iterator

    def add(self, typename, schema):
        self.families[typename] = schema

    def __call__(self, root_name, params):
        result = {}
        for name, typename, v in self.iterator(params):
            mapper = self.families[typename]
            result[name] = mapper(name, v)
        return self.aggregate(root_name, result)


# # sample

# @t.as_schema
# class FieldParams(object):
#     description = t.column(t.TextField)
#     choices = t.column(t.Field)  # xxx
#     widget = t.column(t.TextField, post=t.OneOf(["select", "checkbox", "radio", "multiselect"]))
#     required = t.column(t.BooleanField, default=False, required=False)


# def create_field(name, schema, validated):
#     return t.column(t.Field,
#                     post=t.OneOf([x[0] for x in validated["choices"]]),
#                     name=name,
#                     label=validated["description"],
#                     choices=validated["choices"],
#                     widget=validated["widget"],
#                     required=validated["required"])

import tinyschema as t


def create_schema(name, attrs):
    return t.as_schema(type(name, (object, ), attrs))



import pytest


def _getTarget():
    from tinyschema import Schema
    return Schema


def _makeOne(*args, **kwargs):
    import tinyschema as t

    class S(_getTarget()):
        x = t.column(t.IntegerField, t.positive, x="yyy")
    return S(*args, **kwargs)


def test_any_field_required_None__is_first_converter():
    from tinyschema import reject_None
    s = _makeOne(x="yyy")
    assert s.x.convertors[0] == reject_None


def test_option_access_after_initialized():
    s = _makeOne(x="yyy")
    assert s.x.x == "yyy"


def test_option_access_after_bind():
    s = _makeOne(x="yyy")
    s.x = s.x.bind(x="zzz")
    assert s.x.x == "zzz"


def test_converter_access_after_initialized():
    import tinyschema as t
    s = _makeOne(x=10)
    assert t.positive in s.x.convertors


def test_converter_access_after_bind():
    import tinyschema as t
    s = _makeOne(x=10)
    s.x = s.x.bind(t.OneOf(["a", "b", "c"]))
    assert t.positive in s.x.convertors
    assert any(isinstance(e, t.OneOf) for e in s.x.convertors)


def test_schema_validation__success():
    s = _makeOne(x="10")
    result = s.validate()
    assert result == {"x": 10}


def test_schema_validation__failure():
    import tinyschema as t
    s = _makeOne(x="aa")

    with pytest.raises(t.Failure) as e:
        s.validate()
    assert bool(e.value.errors) is True



# -*- coding:utf-8 -*-


def _callFUT(*args, **kwargs):
    from tinyschema.parser import from_multidict
    return from_multidict(*args, **kwargs)


def test_dict__return_onesself():
    mdict = {"name": "foo", "country": "jp"}
    result = _callFUT(mdict)

    assert result == {"name": "foo", "country": "jp"}


def test_single_mdict__return_dict():
    from webob.multidict import MultiDict

    mdict = MultiDict({"name": "foo", "country": "jp"})
    result = _callFUT(mdict)

    assert result == {"name": "foo", "country": "jp"}


def test_django_like_mdict__return_dict():
    from webob.multidict import MultiDict
    from tinyschema.parser import DjangoMultiDictWrapper

    mdict = DjangoMultiDictWrapper(MultiDict({"name": "foo", "country": "jp"}))
    result = _callFUT(mdict)

    assert result == {"name": "foo", "country": "jp"}


def test_multiple__convert_to_dict_list__order_also_same():
    from webob.multidict import MultiDict

    mdict = MultiDict([
        ("name", "foo"), ("country", "jp"),
        ("name", "bar"), ("country", "us"),
        ("name", "boo"), ("country", "ch"),
    ])
    result = _callFUT(mdict)
    assert result == [{"name": "foo", "country": "jp"},
                      {"name": "bar", "country": "us"},
                      {"name": "boo", "country": "ch"}]


def test_multiple2__convert_to_dict_list__order_also_same():
    from webob.multidict import MultiDict

    mdict = MultiDict([
        ("name", "foo"), ("name", "bar"), ("name", "boo"),
        ("country", "jp"), ("country", "us"), ("country", "ch"),
    ])
    result = _callFUT(mdict)
    assert result == [{"name": "foo", "country": "jp"},
                      {"name": "bar", "country": "us"},
                      {"name": "boo", "country": "ch"}]


def test_php_compatible_mdict_return_list():
    from webob.multidict import MultiDict

    mdict = MultiDict([("name", "foo"), ("g_id[]", "1"), ("g_id[]", "2")])
    result = _callFUT(mdict)
    assert result == {"name": "foo", "g_id": ["1", "2"]}



# -*- coding:utf-8 -*-
def setup_module(m):
    from tinyschema import set_translator, get_translator
    prev = get_translator()
    set_translator(languages=["ja"])
    this = get_translator()
    assert id(prev) != id(this)


def teardown_module(m):
    from tinyschema import set_translator
    set_translator(languages=None)


def test_positive():
    from tinyschema import Failure, Schema, column, PositiveIntegerField
    import pytest

    class S(Schema):
        value = column(PositiveIntegerField)

    with pytest.raises(Failure) as e:
        S(value=-10).validate()
    assert "小さいです" in e.value.errors["value"][0]



# -*- coding:utf-8 -*-
from . import (
    _Field,
    _Collection,
    _Container
)


type_dict = {
    _Field: "atom",
    _Collection: "array",
    _Container: "object"
}


def subschema_of(field):
    # PartialApplicationLike(partial(_Container, schema), required=True).partial
    return field.func.args[0]


def original_of(fn):
    if hasattr(fn, "func"):
        return original_of(fn.func)
    return fn


def type_of(cls):
    if not hasattr(cls, "__mro__"):
        return type_dict[original_of(cls)]
    mro = cls.__mro__
    if mro[1] is object:
        return "object"
    else:
        raise Exception(cls)



# -*- coding:utf-8 -*-
from mako.lookup import TemplateLookup
import os.path

here = os.path.abspath(os.path.dirname(__file__))


class SchemaRenderer(object):
    def __init__(self, template_name="form.mako", directories=[here], lookup=TemplateLookup):
        self.template_name = template_name
        self.lookup = lookup(directories)
        self.template = None
        self.cache = {}

    def renderer(self, widget_name):
        try:
            return self.cache[widget_name]
        except KeyError:
            if self.template is None:
                self.template = self.lookup.get_template(self.template_name)
            renderer = self.template.get_def(widget_name)
            self.cache[widget_name] = renderer
            return renderer

    def render_schema(self, schema):
        r = []
        for field in schema:
            r.append(self.render_field(field))
        return u"".join(r)

    def render_field(self, field):
        return self.renderer(field.widget).render(field)



# -*- coding:utf-8 -*-
import logging
logger = logging.getLogger(__name__)
import re
from functools import partial
from collections import (
    defaultdict,
    OrderedDict
)
from .langhelpers import gensym
from .compat import (
    text_,
    string_types,
    FileNotFoundError
)
import gettext
import translationstring
import os.path
here = os.path.abspath(os.path.dirname(__file__))
translation = None


def create_translator(languages=None):
    try:
        translation = gettext.translation("tinyschema", os.path.join(here, "locales"), languages=languages, codeset="utf8")
    except FileNotFoundError:
        logger.warn("languages %s is not found", languages)
        translation = None
    return translationstring.Translator(translation)


def set_translator(languages=None):
    global translator
    translator = create_translator(languages)


def get_translator():
    return translator

translator = None
set_translator(None)
_ = translationstring.TranslationStringFactory('tinyschema')


class ValidationError(Exception):
    def __init__(self, error=None, name=None, names=None, message=None):
        self.name = name
        self.error = error
        self.names = names or [name]
        self.message = message


class Failure(Exception):
    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return "<Failure errors={!r}>".format(self.errors)


class Break(Exception):
    def __init__(self, value):
        self.value = value


class _Field(object):
    def __init__(self, value, convertors, options):
        self.value = value
        self.convertors = convertors
        self.options = options

    def __getattr__(self, name):
        try:
            return self.options[name]
        except KeyError as e:
            raise AttributeError(e)

    def bind(self, *new_convertors, **options):
        convertors = self.convertors[:]
        convertors.extend(new_convertors)
        new_options = self.options.copy()
        new_options.update(options)
        return self.__class__(self.value, convertors, new_options)

    def validate(self):
        try:
            value = self.value
            for cnv in self.convertors:
                value = cnv(value, self.options)
            return value
        except Break as e:
            return e.value
        except ValidationError as e:
            e.name = self.name
            raise
        except Exception as e:
            raise ValidationError(e, name=self.name)

    def renewal(self, value):
        self.value = value


class _Container(object):
    def __init__(self, schema, value, convertors, options):
        if isinstance(value, schema):
            self.value = value
        else:
            self.value = schema.fromdict(value)
        self.schema = schema
        self.convertors = convertors
        self.options = options

    def __getattr__(self, k):
        try:
            return getattr(self.value, k)
        except AttributeError:
            return self.options[k]

    def __repr__(self):
        return "<Container {!r}>".format(self.value.__class__)

    def renewal(self, value):
        pass


class _Collection(object):
    def __init__(self, schema, values, convertors, options):
        self.value = []
        for v in values:
            if isinstance(v, schema):
                self.value.append(v)
            else:
                self.value.append(schema.fromdict(v))
        self.schema = schema
        self.convertors = convertors
        self.options = options

    def __getattr__(self, k):
        return self.options[k]

    def __getitem__(self, i):
        return self.value[i]

    def __repr__(self):
        return "<Collection {!r}>".format(self.value.__class__)

    def renewal(self, value):
        pass

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def validate(self):
        return [v.validate() for v in self.value]


class PartialApplicationLike(object):
    def __init__(self, func, *args, **options):
        self.func = func
        self.args = list(args)
        self.options = options

    def partial(self, post=None, pre=None, **options_):
        convertors = self.args[:]
        if pre:
            convertors.insert(0, pre)
        if post:
            convertors.append(post)
        options = self.options.copy()
        options.update(options_)
        return PartialApplicationLike(self.func, *convertors, **options)

    def __call__(self, value, **options):
        new_options = self.options.copy()
        new_options.update(options)
        return self.func(value, self.args, new_options)


def column(func, *args, **kwargs):
    v = func(*args, **kwargs)
    v._column_counter = gensym()
    return v


def as_schema(cls):
    xs = []
    s = set()
    for c in cls.__mro__:
        for name, attr in c.__dict__.items():
            if hasattr(attr, "_column_counter"):
                if name not in s:
                    xs.append((attr._column_counter, name))
                    s.add(name)

    cls.fieldnames = [name for (_, name) in sorted(xs, key=lambda xs: xs[0])]

    # iter
    if not hasattr(cls, "__iter__"):
        def __iter__(self):
            for name in self.fieldnames:
                yield getattr(self, name)
        cls.__iter__ = __iter__

    if not hasattr(cls, "fromdict"):
        @classmethod
        def fromdict(cls, params):
            kwargs = {}
            for name in cls.fieldnames:
                if name in params:
                    kwargs[name] = params.pop(name)
            ob = cls(**kwargs)
            ob._kwargs = kwargs
            return ob
        cls.fromdict = fromdict

    # validate
    if not hasattr(cls, "_validate"):
        def _validate(self):
            errors = defaultdict(list)
            values = OrderedDict()
            for name in self.fieldnames:
                try:
                    current = getattr(self, name)
                    current.raw = current.value  # xxx
                    validated = current.validate()
                    values[name] = validated
                    current.renewal(validated)
                except ValidationError as e:
                    if e.message is not None:
                        if isinstance(e.message, translationstring.TranslationString):
                            errors[e.name].append(get_translator()(e.message))
                        else:
                            errors[e.name].append(e.message)
                    else:
                        errors[e.name].append(e.error)
            if errors:
                raise Failure(errors=errors)
            return values
        cls._validate = _validate

    if not hasattr(cls, "validate"):
        def validate(self):
            return self._validate()
        cls.validate = validate

    # init
    template = """\
def __init__(self, {kwargs}):
    cls = self.__class__
    {body}
    """
    kwargs = ", ".join(["{}=None".format(name) for name in cls.fieldnames])
    body = "\n    ".join("self.{x} = getattr(cls, '{x}')({x}, name='{x}')".format(x=name) for name in cls.fieldnames)
    exec(template.format(kwargs=kwargs, body=body))
    cls.__init__ = locals()["__init__"]
    return cls


class SchemaMeta(type):
    def __new__(self, name, bases, attrs):
        cls = super(SchemaMeta, self).__new__(self, name, bases, attrs)
        return as_schema(cls)

Schema = SchemaMeta("Schema", (object, ), {})


# field validation
class Any(object):
    def __init__(self, validators):
        self.validators = validators

    def __call__(self, val, options):
        errors = []
        for v in self.validators:
            try:
                val = v(val, options)
                return val
            except Exception:
                errors.append((_("${val} is not validation suceeed.", mapping={"val": val})))
        e = errors[0]
        raise ValidationError(error=e, message=e.message)


class Regex(object):
    def __init__(self, rx, msg=None):
        if isinstance(rx, string_types):
            self.rx = re.compile(rx)
            self.msg = msg
        else:
            self.rx = rx

    def __call__(self, val, options):
        if self.rx.match(val) is None:
            if self.msg:
                msg = self.msg(val)
            msg = _("invalid email ${val}", mapping={"val": val})
            raise ValidationError(message=(msg))
        return val

EMAIL_RE = "(?i)^[A-Z0-9._%!#$%&'*+-/=?^_`{|}~()]+@[A-Z0-9]+([.-][A-Z0-9]+)*\.[A-Z]{2,22}$"
EMail = partial(Regex, EMAIL_RE, lambda val: (_("${val} is not url", mapping={"val": val})))


class Range(object):
    def __init__(self, min, max):
        self.min = min
        self.max = max

    def __call__(self, val, options):
        if self.min > self.val:
            raise ValidationError(message=(_("${val} is smaller than ${min}", mapping={"val": val, "min": self.min})))
        if self.max < self.val:
            raise ValidationError(message=(_("${val} is bigger than ${max}", mapping={"val": val, "max": self.max})))
        return val


class Length(object):
    def __init__(self, min=None, max=None):
        self.min = min
        self.max = max

    def __call__(self, val, options):
        if min and min > self.val:
            raise ValidationError(message=(_("${val} is shorter than ${min}", mapping={"val": val, "min": self.min})))
        if max and max < self.val:
            raise ValidationError(message=(_("${val} is longer than ${max}", mapping={"val": val, "max": self.max})))
        return val


class OneOf(object):
    def __init__(self, choices):
        self.choices = choices

    def __call__(self, val, options):
        if val not in self.choices:
            candidates = u', '.join([text_(x) for x in self.choices])
            raise ValidationError(message=(_("${val} is not in ${candidates}", mapping={"val": val, "candidates": candidates})))
            raise ValueError(self.choices)
        return val


class Subset(object):
    def __init__(self, choices):
        self.choices = choices

    def __call__(self, val, options):
        if not set(val).issubset(self.choices):
            sval = u', '.join([text_(x) for x in val])
            candidates = u', '.join([text_(x) for x in self.choices])
            raise ValidationError(message=(_("${val} is not subset of  ${candidates}", mapping={"val": sval, "candidates": candidates})))
        return val

URL_REGEX = r"""(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’]))"""  # "emacs!

URL = Regex(URL_REGEX)


def reject_None(val, options):
    if options.get("required", True):
        if val is None:
            raise ValidationError(None, message=(_("required")))
        else:
            return val
    elif val is not None:
        return val
    elif "default" in options:
        raise Break(options["default"])
    else:
        raise Break(None)


def parse_int(val, options):
    try:
        return int(val)
    except ValueError as e:
        raise ValidationError(e, message=(_("${val} is not int", mapping={"val": val})))


def parse_bool(val, options):
    try:
        return bool(val)
    except ValueError as e:
        raise ValidationError(e, message=(_("${val} is not bool", mapping={"val": val})))


def parse_float(val, options):
    try:
        return float(val)
    except ValueError as e:
        raise ValidationError(e, message=(_("${val} is not float", mapping={"val": val})))


def parse_text(val, options):
    try:
        return text_(val)
    except ValueError as e:
        raise ValidationError(e, message=(_("${val} is not text", mapping={"val": val})))


def parse_choices(val, options):
    try:
        return [(x, y) for x, y in val]
    except Exception as e:
        raise ValidationError(e, message=(_("${val} is not valid choices", mapping={"val": val})))


def positive(val, option):
    if val < 0:
        raise ValidationError(None, message=(_("${val} is smaller than zero", mapping={"val": val})))
    return val

# SchemaFactory


def Container(schema):
    return PartialApplicationLike(partial(_Container, schema), required=True, container=True).partial


def Collection(schema):
    return PartialApplicationLike(partial(_Collection, schema), required=True, container=True).partial

Field = PartialApplicationLike(_Field, reject_None, required=True).partial
IntegerField = Field(post=parse_int, type="integer").partial
FloatField = Field(post=parse_float, type="number").partial
BooleanField = Field(post=parse_bool, type="boolean").partial
TextField = Field(post=parse_text, type="string").partial
ChoicesField = Field(post=parse_choices, type="choices").partial
PositiveIntegerField = IntegerField(post=positive).partial



# -*- coding:utf-8 -*-
from tinyschema import(
    as_schema,
    PositiveIntegerField,
    column,
    Container,
    Collection,
)


@as_schema
class Point(object):
    x = column(PositiveIntegerField)
    y = column(PositiveIntegerField)
    z = column(PositiveIntegerField, required=False)


@as_schema
class Pair(object):
    l = column(Container(Point), label="l")
    r = column(Container(Point), label="r")


@as_schema
class Plot(object):
    ps = column(Collection(Point))



# -*- coding:utf-8 -*-

import os
import sys

from setuptools import setup, find_packages
here = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(here, 'README.rst')) as f:
        README = f.read()
    with open(os.path.join(here, 'CHANGES.txt')) as f:
        CHANGES = f.read()
except IOError:
    README = CHANGES = ''

install_requires=[
    'setuptools',
    'translationstring'
    ]

docs_extras = [
    ]

tests_require =[
    "pytest"
]
testing_extras = tests_require + [
    ]

from setuptools.command.test import test as TestCommand

class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True
    def run_tests(self):
        import pytest
        pytest.main(self.test_args)


setup(name='tinyschema',
      version='0.2.1',
      description='tiny schema implementation',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        ],
      keywords='',
      author="",
      author_email="",
      url="https://github.com/podhmo/tinyschema",
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires = install_requires,
      extras_require = {
          'testing':testing_extras,
          'docs':docs_extras,
          },
      tests_require = tests_require,
      cmdclass = {'test': PyTest},
      entry_points = """      """
      )



